#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *
from . import GPIO


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_I2C = 0x05
KONASHI_UUID_I2C_CONFIG_GET = "064d0206-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_I2C_DATA = 0x05
KONASHI_UUID_I2C_DATA_IN = "064d0308-8251-49d9-b6f3-f7ba35e5d0a1"


KONASHI_I2C_SDA_PINNB = 6
KONASHI_I2C_SCL_PINNB = 7
class I2CMode(IntEnum):
    STANDARD = 0
    FAST = 1
class I2CConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('mode', c_uint8, 1),
        ('enabled', c_uint8, 1),
        ('', c_uint8, 6)
    ]
    def __init__(self, enable: bool, mode: I2CMode) -> None:
        """I2C configuration.

        Args:
            enable (bool): True to enable, False to disable.
            mode (I2CMode): The I2C mode.
        """
        self.enabled = enable
        self.mode = mode
    def __str__(self):
        s = "KonashiI2CConfig("
        if self.enabled:
            s += "enabled"
        else:
            s += "disabled"
        s += ", mode:"
        if self.mode == I2CMode.STANDARD:
            s += "standard"
        elif self.mode == I2CMode.FAST:
            s += "fast"
        else:
            s += "unknown"
        s += ")"
        return s

class I2COperation(IntEnum):
    WRITE = 0
    READ = 1
    WRITE_READ = 2
class I2CResult(IntEnum):
    DONE = 0  # Transfer completed successfully.
    NACK = 1  # NACK received during transfer.
    BUS_ERR = 2  # Bus error during transfer (misplaced START/STOP).
    ARB_LOST = 3  # Arbitration lost during transfer.
    USAGE_FAULT = 4  # Usage fault.
    SW_FAULT = 5  # SW fault.


class _I2C(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi, gpio) -> None:
        super().__init__(konashi)
        self._gpio = gpio
        self._config = I2CConfig(False, I2CMode.STANDARD)
        self._async_loop = None
        self._data_in_future = None

    def __str__(self):
        return f'KonashiI2C'

    def __repr__(self):
        return f'KonashiI2C()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_I2C_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_I2C_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_I2C_DATA_IN, self._ntf_cb_data_in)


    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = I2CConfig.from_buffer_copy(data)

    def _ntf_cb_data_in(self, sender, data):
        logger.debug("Received input data: {}".format("".join("{:02x}".format(x) for x in data)))
        if self._async_loop is not None and self._data_in_future is not None:
            self._async_loop.call_soon_threadsafe(self._data_in_future.set_result, data)


    async def config(self, config: I2CConfig) -> None:
        """Configure I2C.

        Args:
            config (I2CConfig): The configuration.

        Raises:
            PinUnavailableError: One of the I2C pins is set to another function.
        """
        if config.enabled:
            if self._gpio._config[KONASHI_I2C_SDA_PINNB].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_I2C_SDA_PINNB].function != int(GPIO.GPIOPinFunction.I2C):
                raise PinUnavailableError(f'Pin {KONASHI_I2C_SDA_PINNB} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_I2C_SDA_PINNB].function]}')
            if self._gpio._config[KONASHI_I2C_SCL_PINNB].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_I2C_SCL_PINNB].function != int(GPIO.GPIOPinFunction.I2C):
                raise PinUnavailableError(f'Pin {KONASHI_I2C_SCL_PINNB} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_I2C_SCL_PINNB].function]}')
        b = bytearray([KONASHI_CFG_CMD_I2C]) + bytearray(config)
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_config(self) -> I2CConfig:
        """Get the current I2C configuration

        Returns:
            I2CConfig: The I2C configuration.
        """
        await self._read(KONASHI_UUID_I2C_CONFIG_GET)
        return self._config

    async def transaction(self, operation: I2COperation, address: int, read_len: int, write_data: bytes) -> Tuple[I2CResult, int, bytes]:
        """Perform an I2C transaction.

        Args:
            operation (I2COperation): The transaction operation.
            address (int): The I2C slave address (address range is 0x00 to 0x7F).
            read_len (int): The length of the data to read (0 to 126 bytes).
            write_data (bytes): The data to write (valid length 0 to 124 bytes).

        Returns:
            Tuple[I2CResult, int, bytes]: result, address, bytes.
                I2CResult: The transaction result.
                int: The transaction slave address.
                bytes: The read data (if there is no read data, the length will be 0).

        Raises:
            ValueError: The read length or slave address is out of range, or the write data is too long.
        """
        if read_len > 126:
            raise ValueError("Maximum read length is 126 bytes")
        if address > 0x7F:
            raise ValueError("The I2C address should be in the range [0x01,0x7F]")
        if len(write_data) > 124:
            raise ValueError("Maximum write data length is 124 bytes")
        b = bytearray([KONASHI_CTL_CMD_I2C_DATA, operation, read_len, address]) + bytearray(write_data)
        self._async_loop = asyncio.get_event_loop()
        self._data_in_future = self._async_loop.create_future()
        await self._write(KONASHI_UUID_CONTROL_CMD, b)
        res = await self._data_in_future
        self._async_loop = None
        self._data_in_future = None
        ret = (I2CResult(res[0]), res[1], res[2:])
        return ret



#!/usr/bin/env python3

import asyncio

from . import GPIO
from . import SoftPWM
from . import HardPWM
from . import AIO
from . import I2C
from . import UART
from . import SPI


class _Io:
    def __init__(self, konashi):
        self._gpio = GPIO._GPIO(konashi)
        self._softpwm = SoftPWM._SoftPWM(konashi, self._gpio)
        self._hardpwm = HardPWM._HardPWM(konashi, self._gpio)
        self._analog = AIO._AIO(konashi)
        self._i2c = I2C._I2C(konashi, self._gpio)
        self._uart = UART._UART(konashi)
        self._spi = SPI._SPI(konashi, self._gpio)

    @property
    def gpio(self) -> GPIO._GPIO:
        """This Konashi devices GPIO interface.
        """
        return self._gpio

    @property
    def softpwm(self) -> SoftPWM._SoftPWM:
        """This Konashi devices Software PWM interface.
        """
        return self._softpwm

    @property
    def hardpwm(self) -> HardPWM._HardPWM:
        """This Konashi devices Hardware PWM interface.
        """
        return self._hardpwm

    @property
    def analog(self) -> AIO._AIO:
        """This Konashi devices AIO interface.
        """
        return self._analog

    @property
    def i2c(self) -> I2C._I2C:
        """This Konashi devices I2C interface.
        """
        return self._i2c

    @property
    def uart(self) -> UART._UART:
        """This Konashi devices UART interface.
        """
        return self._uart

    @property
    def spi(self) -> SPI._SPI:
        """This Konashi devices SPI interface.
        """
        return self._spi

    async def _on_connect(self):
        await self._gpio._on_connect()
        await self._softpwm._on_connect()
        await self._hardpwm._on_connect()
        await self._analog._on_connect()
        await self._i2c._on_connect()
        await self._uart._on_connect()
        await self._spi._on_connect()



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


KONASHI_UUID_BUILTIN_PRESSURE = "00002a6d-0000-1000-8000-00805f9b34fb"


class _Pressure(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._cb = None

    def __str__(self):
        return f'KonashiPressure'

    def __repr__(self):
        return f'KonashiPressure()'


    async def _on_connect(self) -> None:
        pass


    def _ntf_cb(self, sender, data):
        d = struct.unpack("<i", data)
        press = d[0]
        press /= 1000
        if self._cb is not None:
            self._cb(press)


    async def set_callback(self, notify_callback: Callable[[float], None]) -> None:
        """Set a callback for the pressure sensor data.

        Args:
            notify_callback (Callable[[float], None]): The callback.
                The function takes 1 parameter and returns nothing:
                float: The pressure in hectopascal.
        """
        if notify_callback is not None:
            self._cb = notify_callback
            await self._enable_notify(KONASHI_UUID_BUILTIN_PRESSURE, self._ntf_cb)
        else:
            await self._disable_notify(KONASHI_UUID_BUILTIN_PRESSURE)
            self._cb = None



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_UART = 0x06
KONASHI_UUID_UART_CONFIG_GET = "064d0207-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_UART_DATA = 0x06
KONASHI_UUID_UART_DATA_IN = "064d0309-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_UUID_UART_DATA_SEND_DONE = "064d030a-8251-49d9-b6f3-f7ba35e5d0a1"


class UARTParity(IntEnum):
    NONE = 0
    ODD = 1
    EVEN = 2
class UARTStopBits(IntEnum):
    HALF = 0
    ONE = 1
    ONEANDAHALF = 2
    TWO = 3
class UARTConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('stop_bits', c_uint8, 2),
        ('parity', c_uint8, 2),
        ('', c_uint8, 3),
        ('enabled', c_uint8, 1),
        ('baudrate', c_uint32)
    ]
    def __init__(self, enable: bool, baudrate: int=0, parity: UARTParity=UARTParity.NONE, stop_bits: UARTStopBits=UARTStopBits.ONE) -> None:
        """UART configuration.
        When enabling, please always spcify the baudrate, parity and stop bits. When disabling, they can be left as default.

        Args:
            enable (bool): True to enable, False to disable.
            baudrate (int, optional): The UART baudrate. Defaults to 0.
            parity (UARTParity, optional): The UART parity. Defaults to UARTParity.NONE.
            stop_bits (UARTStopBits, optional): The UART number of stop bits. Defaults to UARTStopBits.ONE.
        """
        self.enabled = enable
        self.baudrate = baudrate
        self.parity = parity
        self.stop_bits = stop_bits
    def __str__(self):
        s = "KonashiUARTConfig("
        if self.enabled:
            s += "enabled"
        else:
            s += "disabled"
        s += ", "+str(self.baudrate)
        if self.parity == UARTParity.NONE:
            s += ", No parity"
        elif self.parity == UARTParity.ODD:
            s += ", Odd parity"
        elif self.parity == UARTParity.EVEN:
            s += ", Even parity"
        if self.stop_bits == UARTStopBits.HALF:
            s += ", 0.5 stop bits"
        elif self.stop_bits == UARTStopBits.ONE:
            s += ", 1 stop bit"
        elif self.stop_bits == UARTStopBits.ONEANDAHALF:
            s += ", 1.5 stop bits"
        elif self.stop_bits == UARTStopBits.TWO:
            s += ", 2 stop bits"
        s += ")"
        return s


class _UART(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._config = UARTConfig(False, 0, UARTParity.NONE, UARTStopBits.ONE)
        self._async_loop = None
        self._send_done_future = None
        self._data_in_cb = None

    def __str__(self):
        return f'KonashiUART'

    def __repr__(self):
        return f'KonashiUART()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_UART_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_UART_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_UART_DATA_IN, self._ntf_cb_data_in)
        await self._enable_notify(KONASHI_UUID_UART_DATA_SEND_DONE, self._ntf_cb_send_done)


    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = UARTConfig.from_buffer_copy(data)

    def _ntf_cb_data_in(self, sender, data):
        logger.debug("Received input data: {}".format("".join("{:02x}".format(x) for x in data)))
        if self._data_in_cb is not None:
            self._data_in_cb(data)

    def _ntf_cb_send_done(self, sender, data):
        logger.debug("Received send done data: {}".format("".join("{:02x}".format(x) for x in data)))
        if self._async_loop is not None and self._send_done_future is not None:
            self._async_loop.call_soon_threadsafe(self._send_done_future.set_result, data)


    async def config(self, config: UARTConfig) -> None:
        """Configure the UART peripheral.

        Args:
            config (UARTConfig): The configuration.
        """
        b = bytearray([KONASHI_CFG_CMD_UART]) + bytearray(config)
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_config(self) -> UARTConfig:
        """Get the current UART configuration.

        Returns:
            UARTConfig: The UART configuration.
        """
        await self._read(KONASHI_UUID_UART_CONFIG_GET)
        return self._config

    def set_data_in_cb(self, callback: Callable[[bytes], None]) -> None:
        """Set a callback for received UART data.

        Args:
            callback (Callable[[bytes], None]): The callback.
                The function takes 1 parameter and return nothing:
                bytes: The received data.
        """
        self._data_in_cb = callback

    async def send(self, write_data: bytes) -> bool:
        """Send UART data.

        Args:
            write_data (bytes): The data to send (length range is [1,127]).

        Raises:
            ValueError: The write data length is out of range.

        Returns:
            bool: True if successful, False otherwise.
        """
        if len(write_data) == 0:
            raise ValueError("Write data buffer cannot be empty")
        if len(write_data) > 127:
            raise ValueError("Maximum write data length is 127 bytes")
        b = bytearray([KONASHI_CTL_CMD_UART_DATA]) + bytearray(write_data)
        self._async_loop = asyncio.get_event_loop()
        self._send_done_future = self._async_loop.create_future()
        await self._write(KONASHI_UUID_CONTROL_CMD, b)
        res = await self._send_done_future
        self._async_loop = None
        self._send_done_future = None
        if len(res) == 1 and res[0] == 0x01:
            return True
        else:
            return False



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_ANALOG = 0x04
KONASHI_UUID_ANALOG_CONFIG_GET = "064d0205-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_ANALOG = 0x04
KONASHI_UUID_ANALOG_OUTPUT_GET = "064d0306-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_UUID_ANALOG_INPUT = "064d0307-8251-49d9-b6f3-f7ba35e5d0a1"


KONASHI_AIO_COUNT = 3
class ADCRef(IntEnum):
    DISABLE = 0
    REF_1V25 = 0x0+1
    REF_2V5 = 0x1+1
    REF_VDD = 0x2+1
class VDACRef(IntEnum):
    DISABLE = 0
    REF_1V25LN = 0x0+1
    REF_2V5LN = 0x1+1
    REF_1V25 = 0x2+1
    REF_2V5 = 0x3+1
    REF_VDD = 0x4+1
class IDACRange(IntEnum):
    DISABLE = 0
    RANGE0 = 0x0+1  # 0.05~1.6uA range, 50nA step
    RANGE1 = 0x1+1  # 1.6~4.7uA range, 100nA step
    RANGE2 = 0x2+1  # 0.5~16uA range, 500nA step
    RANGE3 = 0x3+1  # 2~64uA range, 2000nA step
class AIOPinDirection(IntEnum):
    INPUT = 0
    OUTPUT = 1
class AIOPinConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('direction', c_uint8, 1),
        ('send_on_change', c_uint8, 1),
        ('', c_uint8, 1),
        ('enabled', c_uint8, 1),
        ('', c_uint8, 4)
    ]
    def __init__(self, enabled: bool, direction: AIOPinDirection=AIOPinDirection.INPUT, send_on_change: bool=True):
        """AIO pin configuration.

        Args:
            enabled (bool): True to enable, False to disable.
            direction (AIOPinDirection, optional): The pin direction. Defaults to AIOPinDirection.INPUT.
            send_on_change (bool, optional): Unused, can be left as default. Defaults to True.
        """
        self.enabled = enabled
        self.send_on_change = send_on_change
        self.direction = direction
class _AIOConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('adc_update_period', c_uint8),
        ('adc_voltage_reference', c_uint8, 4),
        ('', c_uint8, 4),
        ('vdac_voltage_reference', c_uint8, 4),
        ('', c_uint8, 4),
        ('idac_current_step', c_uint8, 4),
        ('', c_uint8, 4)
    ]
class _AIOAllConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('pin', AIOPinConfig*KONASHI_AIO_COUNT),
        ('analog', _AIOConfig),
    ]

class AIOPinControl(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('control_value', c_uint16),
        ('transition_duration', c_uint32)
    ]
    def __init__(self, control_value: int, transition_duration: int=0):
        """AIO pin control.

        Args:
            control_value (int): The output control value. The valid range is [0,65535].
            transition_duration (int, optional): The output value transition duration in milliseconds. The valid range is [0,4294967295]. Defaults to 0.

        Raises:
            ValueError: The control value or transition duration is out of range.1
        """
        if control_value < 0 or control_value > 65535:
            raise ValueError("The valid range for the control value is [0,65535]")
        if transition_duration < 0 or transition_duration > 4294967295:
            raise ValueError("The valid range for the transition duration is [0,4294967295] (unit: 1ms)")
        self.control_value = control_value
        self.transition_duration = transition_duration
    def __str__(self):
        s = "KonashiAIOPinControl("
        if self.control_type is None:
            s += "Control value "+str(self.control_value)+", Transition duration "+str(self.transition_duration)+"ms"
        s += ")"
        return s
class _AIOPinOut(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('valid', c_uint8),
        ('control', AIOPinControl)
    ]
class _AIOPinsOut(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('idac_current_step', c_uint8, 4),
        ('vdac_voltage_reference', c_uint8, 4),
        ('pin', _AIOPinOut*KONASHI_AIO_COUNT)
    ]
class _AIOPinIn(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('valid', c_uint8),
        ('value', c_uint16)
    ]
class _AIOPinsIn(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('adc_voltage_reference', c_uint8, 4),
        ('', c_uint8, 4),
        ('pin', _AIOPinIn*KONASHI_AIO_COUNT)
    ]


class _AIO(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._config = _AIOAllConfig()
        self._output = _AIOPinsOut()
        self._input = _AIOPinsIn()
        self._input_cb = None

    def __str__(self):
        return f'KonashiAIO'

    def __repr__(self):
        return f'KonashiAIO()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_ANALOG_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_ANALOG_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_ANALOG_OUTPUT_GET, self._ntf_cb_output)
        await self._read(KONASHI_UUID_ANALOG_OUTPUT_GET)
        await self._enable_notify(KONASHI_UUID_ANALOG_INPUT, self._ntf_cb_input)
        await self._read(KONASHI_UUID_ANALOG_INPUT)
        

    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = _AIOAllConfig.from_buffer_copy(data)

    def _ntf_cb_output(self, sender, data):
        logger.debug("Received output data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._output = _AIOPinsOut.from_buffer_copy(data)

    def _ntf_cb_input(self, sender, data):
        logger.debug("Received input data: {}".format("".join("{:02x}".format(x) for x in data)))
        _new_input = _AIOPinsIn.from_buffer_copy(data)
        for i in range(KONASHI_AIO_COUNT):
            if _new_input.pin[i].valid:
                if _new_input.pin[i].value != self._input.pin[i].value:
                    if self._input_cb is not None:
                        self._input_cb(i, self._calc_voltage_for_value(_new_input.pin[i].value))
        self._input = _new_input


    def _calc_voltage_for_value(self, value: int) -> float:
        max_ref = None
        if self._config.analog.adc_voltage_reference == ADCRef.DISABLE:
            return None
        elif self._config.analog.adc_voltage_reference == ADCRef.REF_1V25:
            max_ref = 1.25
        elif self._config.analog.adc_voltage_reference == ADCRef.REF_2V5:
            max_ref = 2.5
        elif self._config.analog.adc_voltage_reference == ADCRef.REF_VDD:
            max_ref = 3.3
        else:
            return None
        return value*max_ref/65535

    async def config_adc_period(self, period: float) -> None:
        """Configure the ADC read period.
        The valid range is [0.1,25.6] seconds, in steps of 100ms.

        Args:
            period (float): The read period in seconds.

        Raises:
            ValueError: The period is out of range.
        """
        if period < 0.1 or period > 25.6:
            raise ValueError("Period should be in range [0.1,25.6] seconds")
        val = round(period*10)-1  # period = 100 * (val+1) in ms
        b = bytearray([KONASHI_CFG_CMD_ANALOG, 0xF0, int(val)&0xFF])
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def config_adc_ref(self, ref: ADCRef) -> None:
        """Configure the ADC voltage reference.

        Args:
            ref (ADCRef): The voltage reference.
        """
        b = bytearray([KONASHI_CFG_CMD_ANALOG, 0xE0|(ref&0x0F)])
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def config_vdac_ref(self, ref: VDACRef) -> None:
        """Configure the VDAC voltage reference.

        Args:
            ref (VDACRef): The voltage reference.
        """
        b = bytearray([KONASHI_CFG_CMD_ANALOG, 0xD0|(ref&0x0F)])
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def config_idac_range(self, range: IDACRange) -> None:
        """Configure the IDAC current range.

        Args:
            range (IDACRange): The current range.
        """
        b = bytearray([KONASHI_CFG_CMD_ANALOG, 0xC0|(range&0x0F)])
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_analog_config(self) -> _AIOConfig:
        """Get the analog configuration.
        This holds the configuration of the ADC, VDAC and IDAC.

        Returns:
            AIOConfig: The analog configuration.
                This class has 4 members:
                    adc_update_period: The ADC update period.
                    adc_voltage_reference: The ADC voltage reference.
                    vdac_voltage_reference: The VDAC voltage reference.
                    idac_current_step: The IDAC current step.
        """
        await self._read(KONASHI_UUID_ANALOG_CONFIG_GET)
        return self._config.analog

    async def config_pins(self, configs: Sequence(Tuple[int, AIOPinConfig])) -> None:
        """Configure analog pins.

        Args:
            configs (Sequence[Tuple[int, AIOPinConfig]]): The list of pin configurations.
                For each Tuple:
                int: A bitmask of the pins to apply the configuration to.
                AIOPinConfig: The configuration for the specified pins.
        """
        b = bytearray([KONASHI_CFG_CMD_ANALOG])
        for config in configs:
            for i in range(KONASHI_AIO_COUNT):
                if (config[0]&(1<<i)) > 0:
                    b.extend(bytearray([(i<<4)|(bytes(config[1])[0]&0x0F)]))
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_pins_config(self, pin_bitmask: int) -> List[AIOPinConfig]:
        """Get the configuration of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the configuration for.

        Returns:
            List[AIOPinConfig]: The configurations of the specified pins.
        """
        await self._read(KONASHI_UUID_ANALOG_CONFIG_GET)
        l = []
        for i in range(KONASHI_AIO_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                l.append(self._config.pin[i])
        return l

    def set_input_cb(self, notify_callback: Callable[[int, int], None]) -> None:
        """Set an analog input callback function.

        Args:
            notify_callback (Callable[[int, int], None]): The callback function.
                The function takes 2 parameters and returns nothing:
                int: The input pin number.
                int: The pin input value.
        """
        self._input_cb = notify_callback

    async def control_pins(self, controls: Sequence(Tuple[int, AIOPinControl])) -> None:
        """Control analog pins.

        Args:
            controls (Sequence[Tuple[int, AIOPinControl]]): The list of pin controls.
                For each Tuple:
                int: A bitmask of the pins to apply the control to.
                AIOPinControl: The control for the specified pins.
        """
        b = bytearray([KONASHI_CTL_CMD_ANALOG])
        for control in controls:
            for i in range(KONASHI_AIO_COUNT):
                if (control[0]&(1<<i)) > 0:
                    b.extend(bytearray([i])+bytearray(control[1]))
        await self._write(KONASHI_UUID_CONTROL_CMD, b)

    def calc_control_value_for_voltage(self, voltage: float) -> int:
        """Calculate the control value for the wanted voltage.

        Args:
            voltage (float): The wanted voltage in Volts.

        Raises:
            KonashiDisabledError: The VDAC is disabled.
            KonashiInvalidError: The VDAC configuration is invalid.
            ValueError: The wanted voltage is out of range for the configuration.

        Returns:
            int: The corresponding control value.
        """
        max_ref = None
        if self._config.analog.vdac_voltage_reference == VDACRef.DISABLE:
            raise KonashiDisabledError("The VDAC is not enabled")
        elif self._config.analog.vdac_voltage_reference == VDACRef.REF_1V25LN or self._config.analog.vdac_voltage_reference == VDACRef.REF_1V25:
            max_ref = 1.25
        elif self._config.analog.vdac_voltage_reference == VDACRef.REF_2V5LN or self._config.analog.vdac_voltage_reference == VDACRef.REF_2V5:
            max_ref = 2.5
        elif self._config.analog.vdac_voltage_reference == VDACRef.REF_VDD:
            max_ref = 3.3
        else:
            raise KonashiInvalidError("The VDAC configuration is not valid")
        if voltage > max_ref:
            raise ValueError(f"The target voltage needs to be in the range [0,{max_ref}]")
        return round(voltage*4095/max_ref)

    def calc_control_value_for_current(self, current: float) -> int:
        """Calculate the control value for the wanted current.

        Args:
            current (float): The wanted current in Amperes.

        Raises:
            KonashiDisabledError: The IDAC is disabled.
            KonashiInvalidError: The IDAC configuration is invalid.
            ValueError: The wanted current is out of range for the configuration.

        Returns:
            int: The corresponding control value.
        """
        first = None
        last = None
        if self._config.analog.idac_current_step == IDACRange.DISABLE:
            raise KonashiDisabledError("The IDAC is not enabled")
        elif self._config.analog.idac_current_step == IDACRange.RANGE0:
            first = 0.05
            last = 1.6
        elif self._config.analog.idac_current_step == IDACRange.RANGE1:
            first = 1.6
            last = 4.7
        elif self._config.analog.idac_current_step == IDACRange.RANGE2:
            first = 0.5
            last = 16
        elif self._config.analog.idac_current_step == IDACRange.RANGE3:
            first = 2
            last = 64
        else:
            raise KonashiInvalidError("The IDAC configuration is not valid")
        if not first <= current <= last:
            raise ValueError(f"The target current needs to be in the range [{first},{last}]")
        return round((current-first)*31/(last-first))

    async def get_pins_control(self, pin_bitmask: int) -> List[AIOPinControl]:
        """Get the output control of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the output control for.

        Returns:
            List[AIOPinControl]: The output control of the specified pins.
        """
        await self._read(KONASHI_UUID_ANALOG_OUTPUT_GET)
        l = []
        for i in range(KONASHI_AIO_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                if not self._output.pin[i].valid:
                    l.append(None)
                else:
                    l.append(self._output.pin[i].control)
        return l

    async def read_pins(self, pin_bitmask: int) -> List[int]:
        """Get the input value of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the input value for.

        Returns:
            List[int]: The input value of the specified pins in Volts.
        """
        await self._read(KONASHI_UUID_ANALOG_INPUT)
        l = []
        for i in range(KONASHI_AIO_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                if not self._input.pin[i].valid:
                    l.append(None)
                else:
                    l.append(self._calc_voltage_for_value(self._input.pin[i].value))
        return l



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *
from . import GPIO


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_HARDPWM = 0x03
KONASHI_UUID_HARDPWM_CONFIG_GET = "064d0204-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_HARDPWM = 0x03
KONASHI_UUID_HARDPWM_OUTPUT_GET = "064d0305-8251-49d9-b6f3-f7ba35e5d0a1"


KONASHI_HARDPWM_COUNT = 4
KONASHI_HARDPWM_PIN_TO_GPIO_NUM = [0, 1, 2, 3]
class HardPWMClock(IntEnum):
    HFCLK = 0
    CASCADE = 1
KONASHI_HARDPWM_CLOCK_FREQ = {HardPWMClock.HFCLK: 38400000, HardPWMClock.CASCADE: 20000}
class HardPWMPrescale(IntEnum):
    DIV1 = 0x0
    DIV2 = 0x1
    DIV4 = 0x2
    DIV8 = 0x3
    DIV16 = 0x4
    DIV32 = 0x5
    DIV64 = 0x6
    DIV128 = 0x7
    DIV256 = 0x8
    DIV512 = 0x9
    DIV1024 = 0xA
class _HardPWMPinConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('enabled', c_uint8, 1),
        ('', c_uint8, 7)
    ]
    def __str__(self):
        s = "KonashiHardPWMPinConfig("
        if self.enabled:
            s += "enabled"
        else:
            s += "disabled"
        s += ")"
        return s
class HardPWMConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('prescale', c_uint8, 4),
        ('clock', c_uint8, 4),
        ('top', c_uint16)
    ]
    def __init__(self, clock: HardPWMClock, prescale: HardPWMPrescale, top: int) -> None:
        """The Hardware PWM configuration.

        Args:
            clock (HardPWMClock): The source clock.
            prescale (HardPWMPrescale): The clock prescale (only available for HFCLK clock).
            top (int): The TOP value (the maximum count value of the timer). The valid range is [0,65535].

        Raises:
            ValueError: The top value is out of range.
        """
        self.clock = clock
        self.prescale = prescale
        if top < 0 or top > 65535:
            raise ValueError("The valid range for top is [0,65535]")
        self.top = top
    def __str__(self):
        s = "KonashiHardPWMConfig("
        if self.clock == HardPWMClock.HFCLK:
            s += "Clock=HF"
        elif self.clock == HardPWMClock.CASCADE:
            s += "Clock=CASCADE"
        else:
            s += "Clock=unknown"
        s += ", Prescale={}".format(1<<self.prescale)
        s += ", Top={}".format(self.top)
        s += ", Period={}s".format((self.top/(KONASHI_HARDPWM_CLOCK_FREQ[self.clock]/(1<<self.prescale))))
        s += ")"
        return s
class _Config(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('pin', _HardPWMPinConfig*KONASHI_HARDPWM_COUNT),
        ('pwm', HardPWMConfig)
    ]

class HardPWMPinControl(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('control_value', c_uint16),
        ('transition_duration', c_uint32)
    ]
    def __init__(self, control_value: int, transition_duration: int=0):
        """Hardware PWM pin control.

        Args:
            control_value (int): The output control value. The valid range is [0,65535].
            transition_duration (int, optional): The output value transition duration in milliseconds. The valid range is [0,4294967295]. Defaults to 0.

        Raises:
            ValueError: The control value or transition duration is out of range.1
        """
        if control_value < 0 or control_value > 65535:
            raise ValueError("The valid range for the control value is [0,65535]")
        if transition_duration < 0 or transition_duration > 4294967295:
            raise ValueError("The valid range for the transition duration is [0,4294967295] (unit: 1ms)")
        self.control_value = control_value
        self.transition_duration = transition_duration
    def __str__(self):
        s = "KonashiHardPWMPinControl("
        s += "Control value "+str(self.control_value)+", Transition duration "+str(self.transition_duration)+"ms"
        s += ")"
        return s
_PinsControl = HardPWMPinControl*KONASHI_HARDPWM_COUNT


class _HardPWM(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi, gpio) -> None:
        super().__init__(konashi)
        self._gpio = gpio
        self._config = _Config()
        self._output = _PinsControl()
        self._trans_end_cb = None
        self._ongoing_control = []

    def __str__(self):
        return f'KonashiHardPWM'

    def __repr__(self):
        return f'KonashiHardPWM()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_HARDPWM_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_HARDPWM_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_HARDPWM_OUTPUT_GET, self._ntf_cb_output)
        await self._read(KONASHI_UUID_HARDPWM_OUTPUT_GET)


    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = _Config.from_buffer_copy(data)

    def _ntf_cb_output(self, sender, data):
        logger.debug("Received output data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._output = _PinsControl.from_buffer_copy(data)
        for i in range(KONASHI_HARDPWM_COUNT):
            if i in self._ongoing_control and self._output[i].transition_duration == 0:
                self._ongoing_control.remove(i)
                if self._trans_end_cb is not None:
                    self._trans_end_cb(i, self._calc_duty_for_control_value(self._output[i].control_value))


    def _calc_pwm_config_for_period(self, period: float) -> HardPWMConfig:
        freq = KONASHI_HARDPWM_CLOCK_FREQ[HardPWMClock.HFCLK]
        presc = None
        top = None
        for div in HardPWMPrescale:
            f = freq/(1<<div.value)
            top = round(period*f)
            if top <= 65535:
                presc = div
                break
        if presc is not None:
            return HardPWMConfig(HardPWMClock.HFCLK, presc, top)
        freq = KONASHI_HARDPWM_CLOCK_FREQ[HardPWMClock.CASCADE]
        presc = HardPWMPrescale.DIV1
        top = round(period*freq)
        if top <= 65535:
            return HardPWMConfig(HardPWMClock.CASCADE, presc, top)
        raise ValueError("A suitable configuration cannot be calculated for this period value")

    def _calc_duty_for_control_value(self, value: int) -> float:
        return value * 100.0 / self._config.pwm.top


    async def config_pwm(self, period: float) -> None:
        """Configure the Hardware PWM with the given period.

        Args:
            period (float): The period in seconds.

        Raises:
            ValueError: A suitable configuration could not be found for the given period.
        """
        config = self._calc_pwm_config_for_period(period)
        b = bytearray([KONASHI_CFG_CMD_HARDPWM, 0xFF]) + bytearray(config)
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_pwm_config(self) -> HardPWMConfig:
        """Get the current Hardware PWM configuration.

        Returns:
            HardPWMConfig: The Hardware PWM configuration.
        """
        await self._read(KONASHI_UUID_HARDPWM_CONFIG_GET)
        return self._config.pwm

    async def config_pins(self, configs: Sequence(Tuple[int, bool])) -> None:
        """Configure the specified Hardware PWM pins.

        Args:
            configs (Sequence[Tuple[int, bool]]): The list of pin configurations.
                For each Tuple:
                int: A bitmask of the pins to apply the configuration to.
                bool: True to enable, False to disable the specified pins.

        Raises:
            PinUnavailableError: At least one of the specified pins is already configured with another function.
        """
        b = bytearray([KONASHI_CFG_CMD_HARDPWM])
        for config in configs:
            for i in range(KONASHI_HARDPWM_COUNT):
                if (config[0]&(1<<i)) > 0:
                    if self._gpio._config[KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]].function != int(GPIO.GPIOPinFunction.PWM):
                        raise PinUnavailableError(f'Pin {KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]].function]}')
                    b.extend(bytearray([(i<<4)|int(config[1])]))
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_pins_config(self, pin_bitmask: int) -> List[_HardPWMPinConfig]:
        """Get the configuration of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the configuration for.

        Returns:
            List[HardPWMPinConfig]: The configurations of the specified pins.
                Each class in the list has the following member:
                enabled: True if the pin is enabled, otherwise False

        """
        await self._read(KONASHI_UUID_HARDPWM_CONFIG_GET)
        l = []
        for i in range(KONASHI_HARDPWM_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                l.append(self._config.pin[i])
        return l

    def set_transition_end_cb(self, notify_callback: Callable[[int, float], None]) -> None:
        """Set a transition end callback function.

        Args:
            notify_callback (Callable[[int, float], None]): The callback function.
                The function takes 2 parameters and returns nothing:
                int: The pin number.
                float: The current duty in %.
        """
        self._trans_end_cb = notify_callback

    async def control_pins(self, controls: Sequence(Tuple[int, HardPWMPinControl])) -> None:
        """Control Hardware PWM pins.

        Args:
            controls (Sequence[Tuple[int, HardPWMPinControl]]): A list of pin controls.
                For each Tuple:
                int: A bitmask of the pins to apply the control to.
                HardPWMPinControl: The control for the specified pins.

        Raises:
            PinUnavailableError: At least one pin is not configured as a Hardware PWM pin.
        """
        ongoing_control = []
        b = bytearray([KONASHI_CTL_CMD_HARDPWM])
        for control in controls:
            for i in range(KONASHI_HARDPWM_COUNT):
                if (control[0]&(1<<i)) > 0:
                    if self._gpio._config[KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]].function != int(GPIO.GPIOPinFunction.PWM):
                        raise PinUnavailableError(f'Pin {KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]} is not configured as PWM (configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_HARDPWM_PIN_TO_GPIO_NUM[i]].function]})')
                    b.extend(bytearray([i])+bytearray(control[1]))
                    ongoing_control.append(i)
        await self._write(KONASHI_UUID_CONTROL_CMD, b)
        for i in ongoing_control:
            if i not in self._ongoing_control:
                self._ongoing_control.append(i)

    def calc_control_value_for_duty(self, duty: float) -> int:
        """Calculate the control value for the wanted duty.

        Args:
            duty (float): The wanted duty cycle value in %.

        Returns:
            int: The corresponding control value.
        """
        return round(duty * self._config.pwm.top / 100.0)

    async def get_pins_control(self, pin_bitmask: int) -> List[HardPWMPinControl]:
        """Get the output control of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the output control for.

        Returns:
            List[HardPWMPinControl]: The output control of the specified pins.
        """
        await self._read(KONASHI_UUID_HARDPWM_OUTPUT_GET)
        l = []
        for i in range(KONASHI_HARDPWM_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                l.append(self._output[i])
        return l



#!/usr/bin/env python3

import asyncio

from . import System
from . import Bluetooth


class _Settings:
    def __init__(self, konashi):
        self._system = System._System(konashi)
        self._bluetooth = Bluetooth._Bluetooth(konashi)

    @property
    def system(self) -> System._System:
        """This Konashi devices System Settings interface.
        """
        return self._system

    @property
    def bluetooth(self) -> Bluetooth._Bluetooth:
        """This Konashi devices Bluetooth Settings interface.
        """
        return self._bluetooth

    async def _on_connect(self):
        await self._system._on_connect()
        await self._bluetooth._on_connect()



#!/usr/bin/env python3


from .Konashi import Konashi
from .Konashi import KonashiScanner


from .Settings.System import SystemSettingsNvmUse
from .Settings.System import SystemSettingsNvmSaveTrigger

from .Settings.Bluetooth import BluetoothSettingsFunction
from .Settings.Bluetooth import BluetoothSettingsPrimaryPhy
from .Settings.Bluetooth import BluetoothSettingsSecondaryPhy
from .Settings.Bluetooth import BluetoothSettingsConnectionPhy
from .Settings.Bluetooth import BluetoothSettingsExAdvertiseContents
from .Settings.Bluetooth import BluetoothSettingsExAdvertiseStatus


from .Io.AIO import ADCRef
from .Io.AIO import VDACRef
from .Io.AIO import IDACRange
from .Io.AIO import AIOPinDirection
from .Io.AIO import AIOPinConfig
from .Io.AIO import AIOPinControl

from .Io.GPIO import GPIOPinFunction
from .Io.GPIO import GPIOPinDirection
from .Io.GPIO import GPIOPinPull
from .Io.GPIO import GPIOPinConfig
from .Io.GPIO import GPIOPinControl
from .Io.GPIO import GPIOPinLevel

from .Io.HardPWM import HardPWMClock
from .Io.HardPWM import HardPWMPrescale
from .Io.HardPWM import HardPWMConfig
from .Io.HardPWM import HardPWMPinControl

from .Io.I2C import I2CMode
from .Io.I2C import I2CConfig
from .Io.I2C import I2COperation
from .Io.I2C import I2CResult

from .Io.SoftPWM import SoftPWMControlType
from .Io.SoftPWM import SoftPWMPinConfig
from .Io.SoftPWM import SoftPWMPinControl

from .Io.SPI import SPIMode
from .Io.SPI import SPIEndian
from .Io.SPI import SPIConfig

from .Io.UART import UARTParity
from .Io.UART import UARTStopBits
from .Io.UART import UARTConfig



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


KONASHI_UUID_BUILTIN_RGB_SET = "064d0403-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_UUID_BUILTIN_RGB_GET = "064d0404-8251-49d9-b6f3-f7ba35e5d0a1"


class _RGBLed(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._cb = None

    def __str__(self):
        return f'KonashiRGBLed'

    def __repr__(self):
        return f'KonashiRGBLed()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_BUILTIN_RGB_GET, self._ntf_cb)


    def _ntf_cb(self, sender, data):
        d = struct.unpack("<ccccH", data)
        color = (d[0],d[1],d[2],d[3])
        if self._cb is not None:
            self._cb(color)
            self._cb = None


    async def set(self, r: int, g: int, b: int, a: int, duration: int, callback: Callable[[Tuple[int,int,int,int]], None] = None) -> None:
        """Set the color of the LED, transitioning during the specified duration.
        If a callback is provided, it will be called when the color transition has finished.

        Args:
            r (int): The Red color component (valid range is [0,255], the value is truncated if out of range).
            g (int): The Green color component (valid range is [0,255], the value is truncated if out of range).
            b (int): The Blue color component (valid range is [0,255], the value is truncated if out of range).
            a (int): The Alpha component (valid range is [0,255], the value is truncated if out of range).
            duration (int): The time to transition to the new color in milliseconds (valid range is [0,65535], the value is truncated if out of range).
            callback (Callable[[Tuple[int,int,int,int]], None], optional): The callback. Defaults to None.
                The function takes 1 parameter and returns nothing:
                Tuple[int,int,int,int]: The current LED color in the form (red,green,blue,alpha).
        """
        b = bytearray([r&0xFF, g&0xFF, b&0xFF, a&0xFF, (duration&0x00FF), ((duration&0xFF00)>>8)])
        await self._write(KONASHI_UUID_BUILTIN_RGB_SET, b)
        if callback is not None:
            self._cb = callback



#!/usr/bin/env python3


class KonashiError(Exception):
    pass

class KonashiConnectionError(Exception):
    pass

class KonashiInvalidError(Exception):
    pass

class KonashiDisabledError(Exception):
    pass

class NotFoundError(Exception):
    pass

class InvalidDeviceError(Exception):
    pass

class PinInvalidError(Exception):
    pass

class PinUnavailableError(Exception):
    pass




#!/usr/bin/env python3

import asyncio
import logging
from ctypes import *

from .. import KonashiElementBase
from . import Temperature
from . import Humidity
from . import Pressure
from . import Presence
from . import AccelGyro
from . import RGBLed


logger = logging.getLogger(__name__)


KONASHI_UUID_BUILTIN_VERSION = "064d0401-8251-49d9-b6f3-f7ba35e5d0a1"


class _Builtin(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi):
        super().__init__(konashi)
        self._temperature = Temperature._Temperature(konashi)
        self._humidity = Humidity._Humidity(konashi)
        self._pressure = Pressure._Pressure(konashi)
        self._presence = Presence._Presence(konashi)
        self._accelgyro = AccelGyro._AccelGyro(konashi)
        self._rgbled = RGBLed._RGBLed(konashi)
        self._version = -1

    @property
    def temperature(self) -> Temperature._Temperature:
        """This Konashi devices temperature sensor interface.
        """
        return self._temperature

    @property
    def humidity(self) -> Humidity._Humidity:
        """This Konashi devices humidity sendor interface.
        """
        return self._humidity

    @property
    def pressure(self) -> Pressure._Pressure:
        """This Konashi devices pressure sensor interface.
        """
        return self._pressure

    @property
    def presence(self) -> Presence._Presence:
        """This Konashi devices presence sensor interface.
        """
        return self._presence

    @property
    def accelgyro(self) -> AccelGyro._AccelGyro:
        """This Konashi devices Accel and Gyro sensor interface.
        """
        return self._accelgyro

    @property
    def rgbled(self) -> RGBLed._RGBLed:
        """This Konashi devices RGB LED interface.
        """
        return self._rgbled

    async def _on_connect(self):
        await self._enable_notify(KONASHI_UUID_BUILTIN_VERSION, self._ntf_cb_version)
        await self._read(KONASHI_UUID_BUILTIN_VERSION)
        await self._temperature._on_connect()
        await self._humidity._on_connect()
        await self._pressure._on_connect()
        await self._presence._on_connect()
        await self._accelgyro._on_connect()
        await self._rgbled._on_connect()

    def _ntf_cb_version(self, sender, data):
        logger.debug("Received konashi built-in version: {}".format("".join("{:02x}".format(x) for x in data)))
        new_version = int.from_bytes(data, byteorder='little', signed=True)
        if new_version != self._version:
            logger.debug("Konashi Built-in version change: {} -> {}".format(self._version, new_version))
            self._version = new_version
            if self._version == 0:
                logger.warn("Konashi base board unknown version")
            elif self._version == 1:
                logger.info("Konashi base board connected")
            else:
                logger.info("No base board connected")



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


logger = logging.getLogger(__name__)


KONASHI_UUID_SETTINGS_CMD = "064d0101-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_SET_CMD_BLUETOOTH = 0x02
KONASHI_UUID_BLUETOOTH_SETTINGS_GET = "064d0103-8251-49d9-b6f3-f7ba35e5d0a1"


class BluetoothSettingsFunction(IntEnum):
    MESH = 0
    EX_ADVERTISER = 1
class BluetoothSettingsPrimaryPhy(IntEnum):
    ONE_M_PHY = 0x01
    CODED_PHY = 0x04
class BluetoothSettingsSecondaryPhy(IntEnum):
    ONE_M_PHY = 0x01
    TWO_M_PHY = 0x02
    CODED_PHY = 0x04
class BluetoothSettingsConnectionPhy(IntFlag):
    ONE_M_PHY_UNCODED  = 0x01
    TWO_M_PHY_UNCODED  = 0x02
    CODED_PHY_125k     = 0x04
    CODED_PHY_500k     = 0x08
class BluetoothSettingsExAdvertiseContents(IntFlag):
    NONE         = 0x00000000
    DEVICE_NAME  = 0x01000000
    UUID128      = 0x02000000
    MANUF_DATA   = 0x04000000
    BLE_ALL      = 0x07000000
    GPIO7_IN     = 0x00800000
    GPIO6_IN     = 0x00400000
    GPIO5_IN     = 0x00200000
    GPIO4_IN     = 0x00100000
    GPIO3_IN     = 0x00080000
    GPIO2_IN     = 0x00040000
    GPIO1_IN     = 0x00020000
    GPIO0_IN     = 0x00010000
    GPIO_IN_ALL  = 0x00FF0000
    AIO2_IN      = 0x00004000
    AIO1_IN      = 0x00002000
    AIO0_IN      = 0x00001000
    AIO_IN_ALL   = 0x00007000
class BluetoothSettingsExAdvertiseStatus(IntEnum):
    DISABLED  = 0x0
    LEGACY    = 0x1
    EXTENDED  = 0x2
    ERROR     = 0xF
class _BluetoothSettings(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('enabled_functions', c_uint8),
        ('main_conn_pref_phy', c_uint8, 4),
        ('main_adv_sec_phy', c_uint8, 4),
        ('ex_adv_sec_phy', c_uint8, 4),
        ('ex_adv_prim_phy', c_uint8, 4),
        ('ex_adv_contents', c_uint32, 28),
        ('ex_adv_status', c_uint32, 4)
    ]
    def __str__(self):
        s = "KonashiSettingsBluetoothSettings("
        s += "EN=0x{:02x}".format(self.enabled_functions)
        s += ", MainAdvSecPHY=0x{:01x}".format(self.main_adv_sec_phy)
        s += ", MainConnPHY=0x{:01x}".format(self.main_conn_pref_phy)
        s += ", ExAdvPrimPHY=0x{:01x}".format(self.ex_adv_prim_phy)
        s += ", ExAdvSecPHY=0x{:01x}".format(self.ex_adv_sec_phy)
        s += ", ExAdvStatus=0x{:01x}".format(self.ex_adv_status)
        s += ", ExAdvContents=0x{:07x}".format(self.ex_adv_contents)
        s += ")"
        return s


class _Bluetooth(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._settings: _BluetoothSettings = _BluetoothSettings()

    def __str__(self):
        s = "KonashiSettingsBluetooth("
        s += "EN=0x{:02x}".format(self.enabled_functions)
        s += ", MainAdvSecPHY=0x{:01x}".format(self.main_adv_sec_phy)
        s += ", MainConnPHY=0x{:01x}".format(self.main_conn_pref_phy)
        s += ", ExAdvPrimPHY=0x{:01x}".format(self.ex_adv_prim_phy)
        s += ", ExAdvSecPHY=0x{:01x}".format(self.ex_adv_sec_phy)
        s += ", ExAdvStatus=0x{:01x}".format(self.ex_adv_status)
        s += ", ExAdvContents=0x{:07x}".format(self.ex_adv_contents)
        s += ")"
        return s

    def __repr__(self):
        s = "KonashiSettingsBluetooth("
        s += "EN=0x{:02x}".format(self.enabled_functions)
        s += ", MainAdvSecPHY=0x{:01x}".format(self.main_adv_sec_phy)
        s += ", MainConnPHY=0x{:01x}".format(self.main_conn_pref_phy)
        s += ", ExAdvPrimPHY=0x{:01x}".format(self.ex_adv_prim_phy)
        s += ", ExAdvSecPHY=0x{:01x}".format(self.ex_adv_sec_phy)
        s += ", ExAdvStatus=0x{:01x}".format(self.ex_adv_status)
        s += ", ExAdvContents=0x{:07x}".format(self.ex_adv_contents)
        s += ")"
        return s


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_BLUETOOTH_SETTINGS_GET, self._ntf_cb_settings)
        await self._read(KONASHI_UUID_BLUETOOTH_SETTINGS_GET)
        

    def _ntf_cb_settings(self, sender, data):
        logger.debug("Received settings data: {}".format("".join("{:02x}".format(x) for x in data)))
        data[3:7] = data[-1:]+data[-2:-1]+data[-3:-2]+data[-4:-3]
        self._settings = _BluetoothSettings.from_buffer_copy(data)


    async def get_settings(self) -> _BluetoothSettings:
        """Get the current Bluetooth settings.

        Returns:
            _BluetoothSettings: The current bluetooth settings.
                This class has 7 members:
                    enabled_functions: A bitmask of enabled Bluetooth functions, see ``BluetoothSettingsFunction``.
                    main_conn_pref_phy: The main connection preferred PHY, see ``BluetoothSettingsConnectionPhy``.
                    main_adv_sec_phy: The main advertising secondary PHY, see ``BluetoothSettingsSecondaryPhy``.
                    ex_adv_sec_phy: The secondary advertiser secondary PHY, see ``BluetoothSettingsSecondaryPhy``.
                    ex_adv_prim_phy: The secondary advertiser primary PHY, see ``BluetoothSettingsPrimaryPhy``.
                    ex_adv_contents: A bitmask of the secondary advertiser contents, see ``BluetoothSettingsExAdvertiseContents``.
                    ex_adv_status: The secondary advertiser status, see ``BluetoothSettingsExAdvertiseStatus``.
        """
        await self._read(KONASHI_UUID_BLUETOOTH_SETTINGS_GET)
        return self._settings

    async def enable_function(self, function: BluetoothSettingsFunction, enable: bool) -> None:
        """Enable or disable a Bluetooth functionality.

        Args:
            function (BluetoothSettingsFunction): The function to enable or disable.
            enable (bool): True to enable, False to disable.
        """
        b = bytearray([KONASHI_SET_CMD_BLUETOOTH, (function<<4)+enable])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def set_main_adv_sec_phy(self, phy: BluetoothSettingsSecondaryPhy) -> None:
        """Set the main advertiser Secondary PHY.

        Args:
            phy (BluetoothSettingsSecondaryPhy): The secondary PHY type.
        """
        b = bytearray([KONASHI_SET_CMD_BLUETOOTH, 0xF0+phy])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def set_main_preferred_conn_phy(self, phy: BluetoothSettingsConnectionPhy) -> None:
        """Set the main preferred connection PHYs.
        Multiple preferred PHYs can be set in the form of a bitmask.

        Args:
            phy (BluetoothSettingsConnectionPhy): The preferred connection PHYs
        """
        b = bytearray([KONASHI_SET_CMD_BLUETOOTH, 0xE0+phy])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def set_ex_adv_phy(self, prim_phy: BluetoothSettingsPrimaryPhy, sec_phy: BluetoothSettingsSecondaryPhy) -> None:
        """Set the secondary advertiser primary and secondary PHYs.

        Args:
            prim_phy (BluetoothSettingsPrimaryPhy): The primary PHY.
            sec_phy (BluetoothSettingsSecondaryPhy): The secondary PHY.
        """
        b = bytearray([KONASHI_SET_CMD_BLUETOOTH, 0xD0+prim_phy, 0xC0+sec_phy])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def set_ex_adv_contents(self, contents: BluetoothSettingsExAdvertiseContents) -> None:
        """Set the secondary advertiser advertising contents.
        If the resulting advertising data length is longer than 31 bytes,
        advertising will automatically be in extended mode, otherwise it will be legacy mode.

        Args:
            contents (BluetoothSettingsExAdvertiseContents): A mask of the contents to show.
        """
        b = bytearray([KONASHI_SET_CMD_BLUETOOTH, 0xB0+((contents>>24)&0x0F), (contents>>16)&0xFF, (contents>>8)&0xFF, (contents>>0)&0xFF])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *
import abc

from bleak import *
from bleak.exc import BleakDBusError

from .Errors import *


logger = logging.getLogger(__name__)


class _KonashiElementBase:
    def __init__(self, konashi):
        self._konashi = konashi

    async def _read(self, uuid: str) -> None:
        if self._konashi._ble_client is None:
            raise KonashiConnectionError(f'Connection is not established')
        try:
            logger.debug("Read from {}".format(uuid))
            while True:
                try:
                    await self._konashi._ble_client.read_gatt_char(uuid)
                    break
                except BleakDBusError as e:
                    if e.dbus_error == "org.bluez.Error.InProgress":
                        continue
                    else:
                        raise e
        except BleakError as e:
            raise KonashiError(f'Error occured during BLE read: "{str(e)}"')

    async def _write(self, uuid: str, data: bytes) -> None:
        if self._konashi._ble_client is None:
            raise KonashiConnectionError(f'Connection is not established')
        try:
            logger.debug("Write to {}: {}".format(uuid, "".join("{:02x}".format(x) for x in data)))
            while True:
                try:
                    await self._konashi._ble_client.write_gatt_char(uuid, data, True)
                    break
                except BleakDBusError as e:
                    if e.dbus_error == "org.bluez.Error.InProgress":
                        continue
                    else:
                        raise e
        except BleakError as e:
            raise KonashiError(f'Error occured during BLE write: "{str(e)}"')

    async def _enable_notify(self, uuid: str, cb: Callable[[int, bytearray], None]) -> None:
        if self._konashi._ble_client is None:
            raise KonashiConnectionError(f'Connection is not established')
        try:
            logger.debug("Enable notify for {}".format(uuid))
            while True:
                try:
                    await self._konashi._ble_client.start_notify(uuid, cb)
                    break
                except BleakDBusError as e:
                    if e.dbus_error == "org.bluez.Error.InProgress":
                        continue
                    else:
                        raise e
        except BleakError as e:
            raise KonashiError(f'Error occured during BLE notify start: "{str(e)}"')

    async def _disable_notify(self, uuid: str) -> None:
        if self._konashi._ble_client is None:
            raise KonashiConnectionError(f'Connection is not established')
        try:
            logger.debug("Disable notify for {}".format(uuid))
            while True:
                try:
                    await self._konashi._ble_client.stop_notify(uuid)
                    break
                except BleakDBusError as e:
                    if e.dbus_error == "org.bluez.Error.InProgress":
                        continue
                    else:
                        raise e
        except BleakError as e:
            raise KonashiError(f'Error occured during BLE notify stop: "{str(e)}"')

    @abc.abstractmethod
    async def _on_connect(self) -> None:
        pass



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .Settings import _Settings
from .Io import _Io
from .Builtin import _Builtin
from .Errors import *


logger = logging.getLogger(__name__)


KONASHI_ADV_SERVICE_UUID = "064d0100-8251-49d9-b6f3-f7ba35e5d0a1"


class Konashi:
    """This class represents a single Konashi device.

    Args:
        name (str): The name (BLE advertising name) of this Konashi device.
    """
    def __init__(self, name: str) -> None:
        """Constructor.
        """
        self._name = name
        self._ble_dev = None
        self._ble_client = None
        self._settings: _Settings = _Settings(self)
        self._io: _Io = _Io(self)
        self._builtin: _Builtin = _Builtin(self)

    def __str__(self):
        return f'Konashi {self._name} ({"Unknown" if self._ble_dev is None else self._ble_dev.address})'

    def __repr__(self):
        return f'Konashi(name="{self._name}")'

    def __eq__(self, other):
        if self._ble_dev is not None and other._ble_dev is not None:
            return self._ble_dev.address == other._ble_dev.address
        return self._name == other._name

    def __ne__(self, other):
        return not self.__eq__(other)


    async def connect(self, timeout: float=0.0) -> None:
        """Connect to this Konashi device.

        If the Konashi class instance was created directly by the user and not returned
        from a KonashiScanner, ``KonashiScanner.find()`` will be called internally before the connection
        takes place. In this case, the passed timeout value is also used for ``KonashiScanner.find()``.

        Args:
            timeout (float, optional): The connection timeout in seconds. Defaults to 0.0.

        Raises:
            KonashiConnectionError: The Konashi device was found but the connection failed.
            NotFoundError: The Konashi device was not found within the timeout time.
            InvalidDeviceError: The specified device name was found but it does not appear to be a valid Konashi device.
        """
        if self._ble_dev is None:
            try:
                k = await KonashiScanner.find(self._name, timeout)
                self._ble_dev = k._ble_dev
            except NotFoundError:
                raise
            except InvalidDeviceError:
                raise
        if self._ble_client is None:
            self._ble_client = BleakClient(self._ble_dev.address)
        try:
            logger.debug("Connect to device {}".format(self._name))
            if not timeout > 0.0:
                timeout = None
            _con = await self._ble_client.connect(timeout=timeout)
        except BleakError as e:
            self._ble_client = None
            raise KonashiConnectionError(f'Error occured during BLE connect: "{str(e)}"')
        if _con:
            await self._settings._on_connect()
            await self._io._on_connect()
            await self._builtin._on_connect()

    async def disconnect(self) -> None:
        """Disconnect from this Konashi device.
        """
        if self._ble_client is not None:
            await self._ble_client.disconnect()
            self._ble_client = None

    @property
    def settings(self) -> _Settings:
        """This Konashi devices Settings interface.
        """
        return self._settings

    @property
    def io(self) -> _Io:
        """This Konashi devices I/O interface.
        """
        return self._io

    @property
    def builtin(self) -> _Builtin:
        """This Konashi devices Built-in interface.
        """
        return self._builtin

    @property
    def name(self) -> str:
        """The name (Bluetooth advertising name) of this Konashi device.

        Returns:
            str: The Konashi device name.
        """
        return self._name


class KonashiScanner:
    """This class represents a Konashi scanner.
    """
    def __init__(self) -> None:
        """Constructor.
        """
        self._scanner = None

    @staticmethod
    async def find(name: str, timeout: float=0.0) -> Konashi:
        """Find the Konashi device specified by name. This is a static method.
        If the timeout expires, NotFoundError is raised.
        If the timeout is 0.0, the search continues indefinitely.

        Args:
            name (str): The Konashi device name to search for.
            timeout (float, optional): The search timeout in seconds. Defaults to 0.0.

        Raises:
            NotFoundError: The Konashi device was not found within the timeout time.
            InvalidDeviceError: The specified device name was found but it does not appear to be a valid Konashi device.

        Returns:
            Konashi: The found Konashi device.
        """
        _konashi = None
        _invalid = False
        _scan_task = None
        _scanner = BleakScanner()
        def _scan_cb(dev, adv):
            nonlocal _konashi
            nonlocal _invalid
            if dev.name == name:
                if KONASHI_ADV_SERVICE_UUID in adv.service_uuids:
                    _konashi = Konashi(name)
                    _konashi._ble_dev = dev
                    logger.debug("Found konashi device")
                else:
                    _invalid = True
                _scanner.register_detection_callback(None)
                if _scan_task:
                    _scan_task.cancel()
        _scanner.register_detection_callback(_scan_cb)
        _timedout = False
        async def _scan_coro(t: float) -> None:
            nonlocal _timedout
            try:
                await _scanner.start()
                if t > 0:
                    await asyncio.sleep(t)
                else:
                    while True:
                        await asyncio.sleep(100)
                _timedout = True
            except asyncio.CancelledError:
                _timedout = False
            finally:
                await _scanner.stop()
        logger.debug("Scan for device {} (timeout={}s)".format(name, timeout))
        _scan_task = asyncio.create_task(_scan_coro(timeout))
        await _scan_task
        if _timedout:
            raise NotFoundError(f'Could not find {name}')
        elif _invalid:
            raise InvalidDeviceError(f'{name} is not a Konashi device')
        else:
            return _konashi

    @staticmethod
    async def search(timeout: float=10.0) -> List[Konashi]:
        """Search for Konashi devices.
        The search continues for the specified timeout and a list of found devices is returned.

        Args:
            timeout (float, optional): The search timeout in seconds, has to be longer than 0s. Defaults to 10.0.

        Raises:
            ValueError: The timeout value is invalid.

        Returns:
            List[Konashi]: A list of discovered Konashi devices.
        """
        if not timeout > 0.0:
            raise ValueError("Timeout should be longer than 0 seconds")
        _konashi = []
        def _scan_cb(dev, adv):
            nonlocal _konashi
            if KONASHI_ADV_SERVICE_UUID in adv.service_uuids:
                k = Konashi(dev.name)
                k._ble_dev = dev
                if k not in _konashi:
                    logger.debug("Discovered new konashi: {}".format(dev.name))
                    _konashi.append(k)
        _scanner = BleakScanner()
        _scanner.register_detection_callback(_scan_cb)
        logger.debug("Start searching for konashi")
        await _scanner.start()
        await asyncio.sleep(timeout)
        _scanner.register_detection_callback(None)
        await _scanner.stop()
        logger.debug("Finished searching for konashi")
        return _konashi

    async def scan_start(self, cb: Callable[[Konashi], None]) -> None:
        """Start scanning for Konashi devices.
        The search continues until cancelled and the callback is called when a device is discovered.

        Args:
            cb (Callable[[Konashi], None]): The callback for discovered devices.

        Returns:
            List[Konashi]: A list of discovered Konashi devices.
        """
        if self._scanner is not None:
            return
        def _scan_cb(dev, adv):
            if KONASHI_ADV_SERVICE_UUID in adv.service_uuids:
                k = Konashi(dev.name)
                k._ble_dev = dev
                cb(k)
        self._scanner = BleakScanner()
        self._scanner.register_detection_callback(_scan_cb)
        await self._scanner.start()
        logger.debug("Started scanning for konashi")

    async def scan_stop(self) -> None:
        if self._scanner is None:
            return
        self._scanner.register_detection_callback(None)
        await self._scanner.stop()
        self._scanner = None
        logger.debug("Finished scanning for konashi")

    @property
    def is_scanning(self) -> bool:
        """Indicates if this scanner instance is currently scanning.

        Returns:
            bool: True is the scanner is scanning, otherwise False.
        """
        return self._scanner is not None



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_GPIO = 0x01
KONASHI_UUID_GPIO_CONFIG_GET = "064d0202-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_GPIO= 0x01
KONASHI_UUID_GPIO_OUTPUT_GET = "064d0302-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_UUID_GPIO_INPUT = "064d0303-8251-49d9-b6f3-f7ba35e5d0a1"


KONASHI_GPIO_COUNT = 8
_KONASHI_GPIO_FUNCTION_STR = ["DISABLED", "GPIO", "PWM", "I2C", "SPI"]
class GPIOPinFunction(IntEnum):
    DISABLED = 0
    GPIO = 1
    PWM = 2
    I2C = 3
    SPI = 4
class GPIOPinDirection(IntEnum):
    DISABLED = 0
    INPUT = 1
    OUTPUT = 2
    OPEN_DRAIN = 3
    OPEN_SOURCE = 4
class GPIOPinPull(IntEnum):
    NONE = 0
    UP = 1
    DOWN = 2
class GPIOPinConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('function', c_uint8, 4),
        ('', c_uint8, 4),
        ('pull_down', c_uint8, 1),
        ('pull_up', c_uint8, 1),
        ('wired_fct', c_uint8, 2),
        ('direction', c_uint8, 1),
        ('send_on_change', c_uint8, 1),
        ('', c_uint8, 2)
    ]
    def __init__(self, direction: GPIOPinDirection, pull: GPIOPinPull=GPIOPinPull.NONE, send_on_change: bool=True):
        """GPIO pin configuration.

        Args:
            direction (GPIOPinDirection): The pin direction
            pull (GPIOPinPull, optional): The pin pull direction. Defaults to GPIOPinPull.NONE.
            send_on_change (bool, optional): If True, notify when the pin value changes. Defaults to True.
        """
        if direction == GPIOPinDirection.INPUT:
            self.function = 1
            self.direction = 0
            self.wired_fct = 0
        elif direction == GPIOPinDirection.OUTPUT:
            self.function = 1
            self.direction = 1
            self.wired_fct = 0
        elif direction == GPIOPinDirection.OPEN_DRAIN:
            self.function = 1
            self.direction = 0
            self.wired_fct = 1
        elif direction == GPIOPinDirection.OPEN_SOURCE:
            self.function = 1
            self.direction = 0
            self.wired_fct = 2
        else:
            self.function = 0
            self.direction = 0
            self.wired_fct = 0
        if pull == GPIOPinPull.UP:
            self.pull_down = 0
            self.pull_up = 1
        elif pull == GPIOPinPull.DOWN:
            self.pull_down = 1
            self.pull_up = 0
        else:
            self.pull_down = 0
            self.pull_up = 0
        self.send_on_change = send_on_change
    def __str__(self):
        s = "KonashiGPIOPinConfig("
        try:
            s += _KONASHI_GPIO_FUNCTION_STR[self.function]
            if self.function == GPIOPinFunction.GPIO:
                s += ", "
                s += "OD" if self.wired_fct==1 else "OS" if self.wired_fct==2 else "OUT" if 1 else "IN"
                if self.pull_down:
                    s += ", PDOWN"
                if self.pull_up:
                    s += ", PUP"
                if self.send_on_change:
                    s += ", NTFY"
        except:
            s += ", Unknown"
        s += ")"
        return s
_PinsConfig = GPIOPinConfig*KONASHI_GPIO_COUNT

class GPIOPinControl(IntEnum):
    LOW = 0
    HIGH = 1
    TOGGLE = 2
class GPIOPinLevel(IntEnum):
    LOW = 0
    HIGH = 1
    INVALID = 2
class _PinIO(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('level', c_uint8, 1),
        ('', c_uint8, 3),
        ('valid', c_uint8, 1),
        ('', c_uint8, 3)
    ]
_PinsIO = _PinIO*KONASHI_GPIO_COUNT


class _GPIO(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._config = _PinsConfig()
        self._output = _PinsIO()
        self._input = _PinsIO()
        self._input_cb = None

    def __str__(self):
        return f'KonashiGPIO'

    def __repr__(self):
        return f'KonashiGPIO()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_GPIO_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_GPIO_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_GPIO_OUTPUT_GET, self._ntf_cb_output)
        await self._read(KONASHI_UUID_GPIO_OUTPUT_GET)
        await self._enable_notify(KONASHI_UUID_GPIO_INPUT, self._ntf_cb_input)
        await self._read(KONASHI_UUID_GPIO_INPUT)
        

    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = _PinsConfig.from_buffer_copy(data)

    def _ntf_cb_output(self, sender, data):
        logger.debug("Received output data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._output = _PinsIO.from_buffer_copy(data)

    def _ntf_cb_input(self, sender, data):
        logger.debug("Received input data: {}".format("".join("{:02x}".format(x) for x in data)))
        for i in range(KONASHI_GPIO_COUNT):
            if data[i]&0x10:
                val = data[i]&0x01
                if self._input[i].level != val:
                    if self._input_cb is not None:
                        self._input_cb(i, val)
        self._input = _PinsIO.from_buffer_copy(data)


    async def config_pins(self, configs: Sequence(Tuple[int, GPIOPinConfig])) -> None:
        """Configure GPIO pins.

        Args:
            configs (Sequence[Tuple[int, GPIOPinConfig]]): The list of pin configurations.
                For each Tuple:
                int: A bitmask of the pins to apply the configuration to.
                GPIOPinConfig: The configuration for the specified pins.

        Raises:
            PinUnavailableError: At least one of the specified pins is confgured with a function other than GPIO.
        """
        b = bytearray([KONASHI_CFG_CMD_GPIO])
        for config in configs:
            for i in range(KONASHI_GPIO_COUNT):
                if (config[0]&(1<<i)) > 0:
                    if GPIOPinFunction(self._config[i].function) != GPIOPinFunction.DISABLED and GPIOPinFunction(self._config[i].function) != GPIOPinFunction.GPIO:
                        raise PinUnavailableError(f'Pin {i} is already configured as {_KONASHI_GPIO_FUNCTION_STR[self._config[i].function]}')
                    b.extend(bytearray([(i<<4)|(bytes(config[1])[0]), bytes(config[1])[1]]))
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_pins_config(self, pin_bitmask: int) -> List[GPIOPinConfig]:
        """Get the configuration of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the configuration for.

        Returns:
            List[GPIOPinConfig]: The configurations of the specified pins.
        """
        await self._read(KONASHI_UUID_GPIO_CONFIG_GET)
        l = []
        for i in range(KONASHI_GPIO_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                l.append(self._config[i])
        return l

    def set_input_cb(self, notify_callback: Callable[[int, int], None]) -> None:
        """Set a GPIO input callback function.
        The callback will also be called if the pin is set as open drain or open source.

        Args:
            notify_callback (Callable[[int, int], None]): The callback function.
                The function takes 2 parameters and returns nothing:
                int: The pin number.
                int: The pin value.
        """
        self._input_cb = notify_callback

    async def control_pins(self, controls: Sequence(Tuple[int, GPIOPinControl])) -> None:
        """Control GPIO pins.

        Args:
            controls (Sequence[Tuple[int, GPIOPinControl]]): A list of pin controls.
                For each Tuple:
                int: A bitmask of the pins to apply the control to.
                GPIOPinControl: The control for the specified pins.

        Raises:
            PinUnavailableError: At least one pin is not configured as GPIO.
        """
        b = bytearray([KONASHI_CTL_CMD_GPIO])
        for control in controls:
            for i in range(KONASHI_GPIO_COUNT):
                if (control[0]&(1<<i)) > 0:
                    if GPIOPinFunction(self._config[i].function) != GPIOPinFunction.GPIO:
                        raise PinUnavailableError(f'Pin {i} is not configured as GPIO (configured as {_KONASHI_GPIO_FUNCTION_STR[self._config[i].function]})')
                    b.extend(bytearray([(i<<4)|(control[1])]))
        await self._write(KONASHI_UUID_CONTROL_CMD, b)

    async def get_pins_control(self, pin_bitmask: int) -> List[GPIOPinLevel]:
        """Get the output control of the specified pin.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the output control for.

        Returns:
            List[GPIOPinLevel]: The output control of the specified pins.
        """
        await self._read(KONASHI_UUID_GPIO_OUTPUT_GET)
        l = []
        for i in range(KONASHI_GPIO_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                if not self._output[i].valid:
                    l.append(GPIOPinLevel.INVALID)
                else:
                    l.append(GPIOPinLevel(self._output[i].level))
        return l

    async def read_pins(self, pin_bitmask: int) -> List[GPIOPinLevel]:
        """Get the input value of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the input value for.

        Returns:
            List[GPIOPinLevel]: The input value of the specified pins.
        """
        await self._read(KONASHI_UUID_GPIO_INPUT)
        l = []
        for i in range(KONASHI_GPIO_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                if not self._input[i].valid:
                    l.append(GPIOPinLevel.INVALID)
                else:
                    l.append(GPIOPinLevel(self._input[i].level))
        return l



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


KONASHI_UUID_BUILTIN_ACCELGYRO = "064d0402-8251-49d9-b6f3-f7ba35e5d0a1"


class _AccelGyro(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._cb = None

    def __str__(self):
        return f'KonashiAccelGyro'

    def __repr__(self):
        return f'KonashiAccelGyro()'


    async def _on_connect(self) -> None:
        pass


    def _ntf_cb(self, sender, data):
        d = struct.unpack("<hhhhhh", data)
        accel_x = d[0] / 32768 * 8
        accel_y = d[1] / 32768 * 8
        accel_z = d[2] / 32768 * 8
        gyro_x = d[3] / 32768 * 1000
        gyro_y = d[4] / 32768 * 1000
        gyro_z = d[5] / 32768 * 1000
        if self._cb is not None:
            self._cb((accel_x,accel_y,accel_z),(gyro_x,gyro_y,gyro_z))


    async def set_callback(self, notify_callback: Callable[[Tuple[float,float,float],Tuple[float,float,float]], None]) -> None:
        """Set the callback for the accelerometer and gyroscope data.

        Args:
            notify_callback (Callable[[Tuple[float,float,float],Tuple[float,float,float]], None]): The callback.
                The function take 2 parameters and returns nothing:
                Tuple[float,float,float]: The acceleration in g (9.8 m/s^2).
                Tuple[float,float,float]: The angular speed in degree/s.
        """
        if notify_callback is not None:
            self._cb = notify_callback
            await self._enable_notify(KONASHI_UUID_BUILTIN_ACCELGYRO, self._ntf_cb)
        else:
            await self._disable_notify(KONASHI_UUID_BUILTIN_ACCELGYRO)
            self._cb = None



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *


logger = logging.getLogger(__name__)


KONASHI_UUID_SETTINGS_CMD = "064d0101-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_SET_CMD_SYSTEM = 0x01
KONASHI_UUID_SYSTEM_SETTINGS_GET = "064d0102-8251-49d9-b6f3-f7ba35e5d0a1"


class _Command(IntEnum):
    NVM_USE_SET = 1
    NVM_SAVE_TRIGGER_SET = 2
    NVM_SAVE_NOW = 3
    NVM_ERASE_NOW = 4
    FCT_BTN_EMULATE_PRESS = 5
    FCT_BTN_EMULATE_LONG_PRESS = 6
    FCT_BTN_EMULATE_VERY_LONG_PRESS = 7
class SystemSettingsNvmUse(IntEnum):
    DISABLED = 0  ## Disable NVM use.
    ENABLED = 1  ## Enable NVM use.
class SystemSettingsNvmSaveTrigger(IntEnum):
    AUTO = 0  ## Automatically save to NVM on config change.
    MANUAL = 1  ## Manually save to NVM.
class _SystemSettings(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('nvm_use', c_uint8),
        ('nvm_save_trigger', c_uint8)
    ]
    def __str__(self):
        s = "KonashiSettingsSystemSettings("
        if self.nvm_use == SystemSettingsNvmUse.ENABLED:
            s += "NVM enabled"
            s += ", NVM save " + ("auto" if self.nvm_save_trigger==SystemSettingsNvmSaveTrigger.AUTO else "manual")
        else:
            s += "NVM disabled"
        s += ")"
        return s


class _System(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi) -> None:
        super().__init__(konashi)
        self._settings: _SystemSettings = _SystemSettings()

    def __str__(self):
        s = "KonashiSettingsSystem("
        if self.nvm_use == SystemSettingsNvmUse.ENABLED:
            s += "NVM enabled"
            s += ", NVM save " + ("auto" if self.nvm_save_trigger==SystemSettingsNvmSaveTrigger.AUTO else "manual")
        else:
            s += "NVM disabled"
        s += ")"
        return s

    def __repr__(self):
        return "KonashiSettingsSystem(0x{:02x}, 0x{:02x})".format(self._settings.nvm_use, self._settings.nvm_save_trigger)


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_SYSTEM_SETTINGS_GET, self._ntf_cb_settings)
        await self._read(KONASHI_UUID_SYSTEM_SETTINGS_GET)


    def _ntf_cb_settings(self, sender, data):
        logger.debug("Received settings data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._settings = _SystemSettings.from_buffer_copy(data)


    async def get_settings(self) -> _SystemSettings:
        """Get the current system settings.

        Returns:
            _SystemSettings: The current system settings.
                This class has 2 members:
                    nvm_use: 1 if the NVM is set to be used, 0 otherwise.
                    nvm_save_trigger: 1 if manual save, 0 if auto save.
        """
        await self._read(KONASHI_UUID_SYSTEM_SETTINGS_GET)
        return self._settings

    async def set_nvm_use(self, enable: bool) -> None:
        """Set if NVM is used or not.

        Args:
            enable (bool): True to enable, False to disable.
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.NVM_USE_SET, enable])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def set_nvm_save_trigger(self, trigger: SystemSettingsNvmSaveTrigger) -> None:
        """Set the NVM save trigger.

        Args:
            trigger (SystemSettingsNvmSaveTrigger): ``AUTO`` for automatic save,
                ``MANUAL`` for manual save (``nvm_save_now`` needs to be called to save).
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.NVM_SAVE_TRIGGER_SET, trigger])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def nvm_save_now(self) -> None:
        """Save all to NVM now.
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.NVM_SAVE_NOW])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def nvm_erase_now(self) -> None:
        """Erase all from NVM now.
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.NVM_ERASE_NOW])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def emul_press(self) -> None:
        """Emulate a function button simple press.
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.FCT_BTN_EMULATE_PRESS])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def emul_long_press(self) -> None:
        """Emulate a function button long press.
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.FCT_BTN_EMULATE_LONG_PRESS])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)

    async def emul_very_long_press(self) -> None:
        """Emulate a function button very long press.
        """
        b = bytearray([KONASHI_SET_CMD_SYSTEM, _Command.FCT_BTN_EMULATE_VERY_LONG_PRESS])
        await self._write(KONASHI_UUID_SETTINGS_CMD, b)



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *
from . import GPIO


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_SOFTPWM = 0x02
KONASHI_UUID_SOFTPWM_CONFIG_GET = "064d0203-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_SOFTPWM = 0x02
KONASHI_UUID_SOFTPWM_OUTPUT_GET = "064d0304-8251-49d9-b6f3-f7ba35e5d0a1"


KONASHI_SOFTPWM_COUNT = 4
KONASHI_SOFTPWM_PIN_TO_GPIO_NUM = [4, 5, 6, 7]
class SoftPWMControlType(IntEnum):
    DISABLED = 0
    DUTY = 1
    PERIOD = 2
class SoftPWMPinConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('control_type', c_uint8, 4),
        ('', c_uint8, 4),
        ('fixed_value', c_uint16)
    ]
    def __init__(self, control_type: SoftPWMControlType, fixed_value: int=0):
        """Software PWM pin configuration.

        Args:
            control_type (SoftPWMControlType): The PWM control type.
            fixed_value (int, optional): The fixed value.
                If the control type is ``DUTY``, this is the fixed period in milliseconds (valid range is [0,65535]).
                If the control type is ``PERIDO``, this is the fixed duty cycle in 0.1% (valid range is [0,1000]).
                Defaults to 0.

        Raises:
            ValueError: The fixed value is out of range.
        """
        self.control_type = int(control_type)
        if control_type == SoftPWMControlType.DUTY:
            if fixed_value < 0 or fixed_value > 65535:
                raise ValueError("The valid range for the fixed period is [0,65535] (unit: 1ms)")
        elif control_type == SoftPWMControlType.PERIOD:
            if fixed_value < 0 or fixed_value > 1000:
                raise ValueError("The valid range for the fixed duty cycle is [0,1000] (unit: 0.1%)")
        self.fixed_value = fixed_value
    def __str__(self):
        s = "KonashiSoftPWMPinConfig("
        s += "DUTY control" if self.control_type==SoftPWMControlType.DUTY else "PERIOD control" if self.control_type==SoftPWMControlType.PERIOD else "DISABLED"
        if self.control_type != SoftPWMControlType.DISABLED:
            s += ", "
            s += str(self.fixed_value*0.1)
            if self.control_type == SoftPWMControlType.DUTY:
                s += "ms fixed period"
            else:
                s += r"% fixed duty cycle"
        s += ")"
        return s
_PinsConfig = SoftPWMPinConfig*KONASHI_SOFTPWM_COUNT

class SoftPWMPinControl(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('control_type', c_uint8, 4),
        ('', c_uint8, 4),
        ('control_value', c_uint16),
        ('transition_duration', c_uint32)
    ]
    def __init__(self, control_value: int, transition_duration: int=0):
        """Software PWM pin control.

        Args:
            control_value (int): The control value to set.
                If the control type is ``DUTY``, this is the duty cycle control value in 0.1% (valid range is [0,1000]).
                If the control type id ``PERIOD``, this is the period control value in milliseconds (valid range is [0,65535]).
            transition_duration (int, optional): The transition duration in milliseconds (valid range is [0,4294967295]). Defaults to 0.

        Raises:
            ValueError: The control value or the transition duration is out of range.
        """
        if control_value < 0 or control_value > 65535:
            raise ValueError("The valid range for the control value is [0,65535]")
        if transition_duration < 0 or transition_duration > 4294967295:
            raise ValueError("The valid range for the transition duration is [0,4294967295] (unit: 1ms)")
        self.control_value = control_value
        self.transition_duration = transition_duration
        self.control_type = 0
    def __str__(self):
        s = "KonashiSoftPWMPinControl("
        if self.control_type == SoftPWMControlType.DISABLED:
            s += "DISABLED, Control value "+str(self.control_value)+" raw, Transition duration "+str(self.transition_duration)+"ms"
        else:
            s += str(int(self.control_value*0.1))
            if self.control_type == SoftPWMControlType.PERIOD:
                s += "ms period"
            else:
                s += r"% duty cycle"
            s += ", Transition duration "+str(self.transition_duration)+"ms"
        s += ")"
        return s
_PinsControl = SoftPWMPinControl*KONASHI_SOFTPWM_COUNT


class _SoftPWM(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi, gpio) -> None:
        super().__init__(konashi)
        self._gpio = gpio
        self._config = _PinsConfig()
        self._output = _PinsControl()
        self._trans_end_cb = None
        self._ongoing_control = []

    def __str__(self):
        return f'KonashiSoftPWM'

    def __repr__(self):
        return f'KonashiSoftPWM()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_SOFTPWM_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_SOFTPWM_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_SOFTPWM_OUTPUT_GET, self._ntf_cb_output)
        await self._read(KONASHI_UUID_SOFTPWM_OUTPUT_GET)


    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = _PinsConfig.from_buffer_copy(data)

    def _ntf_cb_output(self, sender, data):
        logger.debug("Received output data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._output = _PinsControl.from_buffer_copy(data)
        for i in range(KONASHI_SOFTPWM_COUNT):
            if i in self._ongoing_control and self._output[i].transition_duration == 0:
                self._ongoing_control.remove(i)
                if self._trans_end_cb is not None:
                    self._trans_end_cb(i, self._output[i].control_type, self._output[i].control_value)


    async def config_pins(self, configs: Sequence(Tuple[int, SoftPWMPinConfig])) -> None:
        """Configure the specified Software PWM pins.

        Args:
            configs (Sequence[Tuple[int, SoftPWMPinConfig]]): The list of pin configurations.
                For each Tuple:
                int: A bitmask of the pins to apply the configuration to.
                SoftPWMPinConfig: The configuration for the specified pins.

        Raises:
            PinUnavailableError: At least one of the specified is already configured with another function.
        """
        b = bytearray([KONASHI_CFG_CMD_SOFTPWM])
        for config in configs:
            for i in range(KONASHI_SOFTPWM_COUNT):
                if (config[0]&(1<<i)) > 0:
                    if self._gpio._config[KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]].function != int(GPIO.GPIOPinFunction.PWM):
                        raise PinUnavailableError(f'Pin {KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]].function]}')
                    b.extend(bytearray([(i<<4)|(bytes(config[1])[0])]) + bytearray(config[1])[1:3])
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_pins_config(self, pin_bitmask: int) -> List[SoftPWMPinConfig]:
        """Get the configuration of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the configuration for.

        Returns:
            List[SoftPWMPinConfig]: The configuration of the specified pins.
        """
        await self._read(KONASHI_UUID_SOFTPWM_CONFIG_GET)
        l = []
        for i in range(KONASHI_SOFTPWM_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                l.append(self._config[i])
        return l

    def set_transition_end_cb(self, notify_callback: Callable[[int, SoftPWMControlType, int], None]) -> None:
        """Set a transition end callback function.

        Args:
            notify_callback (Callable[[int, SoftPWMControlType, int], None]): The callback function.
                The function takes 3 parameters and returns nothing:
                int: The pin number.
                SoftPWMControlType: The pins control type.
                int: The current control value.
        """
        self._trans_end_cb = notify_callback

    async def control_pins(self, controls: Sequence(Tuple[int, SoftPWMPinControl])) -> None:
        """Control Software PWM pins.

        Args:
            controls (Sequence[Tuple[int, SoftPWMPinControl]]): A list of pin controls.
                For each Tuple:
                int: A bitmask of the pins to apply the control to.
                SoftPWMPinControl: The control for the specified pins.

        Raises:
            PinUnavailableError: At least one pin is not configured as a Software PWM pin.
            ValueError: The control value is out of range.
        """
        ongoing_control = []
        b = bytearray([KONASHI_CTL_CMD_SOFTPWM])
        for control in controls:
            for i in range(KONASHI_SOFTPWM_COUNT):
                if (control[0]&(1<<i)) > 0:
                    if self._gpio._config[KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]].function != int(GPIO.GPIOPinFunction.PWM):
                        raise PinUnavailableError(f'Pin {KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]} is not configured as PWM (configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_SOFTPWM_PIN_TO_GPIO_NUM[i]].function]})')
                    if self._config[i].control_type == int(SoftPWMControlType.DUTY):
                        if control[1].control_value < 0 or control[1].control_value > 1000:
                            raise ValueError("The valid range for the duty cycle control is [0,1000] (unit: 0.1%)")
                    elif self._config[i].control_type == int(SoftPWMControlType.PERIOD):
                        if control[1].control_value < 0 or control[1].control_value > 65535:
                            raise ValueError("The valid range for the period control is [0,65535] (unit: 1ms)")
                    else:
                        raise PinUnavailableError(f'SoftPWM{i} is not enabled')
                    b.extend(bytearray([i])+(bytearray(control[1])[1:]))
                    ongoing_control.append(i)
        await self._write(KONASHI_UUID_CONTROL_CMD, b)
        for i in ongoing_control:
            if i not in self._ongoing_control:
                self._ongoing_control.append(i)

    async def get_pins_control(self, pin_bitmask: int) -> List[SoftPWMPinControl]:
        """Get the output control of the specified pins.

        Args:
            pin_bitmask (int): A bitmask of the pins to get the output control for.

        Returns:
            List[SoftPWMPinControl]: The output control of the specified pins.
        """
        await self._read(KONASHI_UUID_SOFTPWM_OUTPUT_GET)
        l = []
        for i in range(KONASHI_SOFTPWM_COUNT):
            if (pin_bitmask&(1<<i)) > 0:
                l.append(self._output[i])
        return l



#!/usr/bin/env python3

from __future__ import annotations

import asyncio
import struct
import logging
from ctypes import *
from typing import *
from enum import *

from bleak import *

from .. import KonashiElementBase
from ..Errors import *
from . import GPIO


logger = logging.getLogger(__name__)


KONASHI_UUID_CONFIG_CMD = "064d0201-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CFG_CMD_SPI = 0x07
KONASHI_UUID_SPI_CONFIG_GET = "064d0208-8251-49d9-b6f3-f7ba35e5d0a1"

KONASHI_UUID_CONTROL_CMD = "064d0301-8251-49d9-b6f3-f7ba35e5d0a1"
KONASHI_CTL_CMD_SPI_DATA = 0x07
KONASHI_UUID_SPI_DATA_IN = "064d030b-8251-49d9-b6f3-f7ba35e5d0a1"


KONASHI_SPI_CS_PINNB = 2
KONASHI_SPI_CLK_PINNB = 5
KONASHI_SPI_MISO_PINNB = 3
KONASHI_SPI_MOSI_PINNB = 4
class SPIMode(IntEnum):
    MODE0 = 0  # SPI mode 0: CLKPOL=0, CLKPHA=0.
    MODE1 = 1  # SPI mode 1: CLKPOL=0, CLKPHA=1.
    MODE2 = 2  # SPI mode 2: CLKPOL=1, CLKPHA=0.
    MODE3 = 3  # SPI mode 3: CLKPOL=1, CLKPHA=1.
class SPIEndian(IntEnum):
    LSB_FIRST = 0  # LSB bit is transmitted first.
    MSB_FIRST = 1  # MSB bit is transmitted first.
class SPIConfig(LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('mode', c_uint8, 2),
        ('', c_uint8, 1),
        ('endian', c_uint8, 1),
        ('', c_uint8, 3),
        ('enabled', c_uint8, 1),
        ('bitrate', c_uint32)
    ]
    def __init__(self, enable: bool, mode: SPIMode=SPIMode.MODE0, endian: SPIEndian=SPIEndian.LSB_FIRST, bitrate: int=0) -> None:
        """SPI configuration.
        When enabling, please always spcify the mode, endian and bitrate. When disabling, they can be left as default.

        Args:
            enable (bool): True to enable, False to disable.
            mode (SPIMode, optional): The SPI mode. Defaults to SPIMode.MODE0.
            endian (SPIEndian, optional): The SPI endiand. Defaults to SPIEndian.LSB_FIRST.
            bitrate (int, optional): The SPI bitrate. Defaults to 0.
        """
        self.enabled = enable
        self.mode = mode
        self.endian = endian
        self.bitrate = bitrate
    def __str__(self):
        s = "KonashiSPIConfig("
        if self.enabled:
            s += "enabled"
        else:
            s += "disabled"
        if self.mode == SPIMode.MODE0:
            s += ", CLKPOL=0/CLKPHA=0"
        elif self.mode == SPIMode.MODE1:
            s += ", CLKPOL=0/CLKPHA=1"
        elif self.mode == SPIMode.MODE2:
            s += ", CLKPOL=1/CLKPHA=0"
        elif self.mode == SPIMode.MODE3:
            s += ", CLKPOL=1/CLKPHA=1"
        if self.endian == SPIEndian.LSB_FIRST:
            s += ", LSB"
        elif self.endian == SPIEndian.MSB_FIRST:
            s += ", MSB"
        s += ", "+str(self.bitrate)
        s += ")"
        return s


class _SPI(KonashiElementBase._KonashiElementBase):
    def __init__(self, konashi, gpio) -> None:
        super().__init__(konashi)
        self._gpio = gpio
        self._config = SPIConfig(False, SPIMode.MODE0, SPIEndian.LSB_FIRST, 0)
        self._async_loop = None
        self._data_in_future = None

    def __str__(self):
        return f'KonashiSpi'

    def __repr__(self):
        return f'KonashiSpi()'


    async def _on_connect(self) -> None:
        await self._enable_notify(KONASHI_UUID_SPI_CONFIG_GET, self._ntf_cb_config)
        await self._read(KONASHI_UUID_SPI_CONFIG_GET)
        await self._enable_notify(KONASHI_UUID_SPI_DATA_IN, self._ntf_cb_data_in)


    def _ntf_cb_config(self, sender, data):
        logger.debug("Received config data: {}".format("".join("{:02x}".format(x) for x in data)))
        self._config = SPIConfig.from_buffer_copy(data)

    def _ntf_cb_data_in(self, sender, data):
        logger.debug("Received input data: {}".format("".join("{:02x}".format(x) for x in data)))
        if self._async_loop is not None and self._data_in_future is not None:
            self._async_loop.call_soon_threadsafe(self._data_in_future.set_result, data)


    async def config(self, config: SPIConfig) -> None:
        """Configure the SPI peripheral. 

        Args:
            config (SPIConfig): The configuration.

        Raises:
            PinUnavailableError: At least one of the pins is already configured with another function.
        """
        if config.enabled:
            if self._gpio._config[KONASHI_SPI_CS_PINNB].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_SPI_CS_PINNB].function != int(GPIO.GPIOPinFunction.SPI):
                raise PinUnavailableError(f'Pin {KONASHI_SPI_CS_PINNB} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_SPI_CS_PINNB].function]}')
            if self._gpio._config[KONASHI_SPI_CLK_PINNB].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_SPI_CLK_PINNB].function != int(GPIO.GPIOPinFunction.SPI):
                raise PinUnavailableError(f'Pin {KONASHI_SPI_CLK_PINNB} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_SPI_CLK_PINNB].function]}')
            if self._gpio._config[KONASHI_SPI_MISO_PINNB].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_SPI_MISO_PINNB].function != int(GPIO.GPIOPinFunction.SPI):
                raise PinUnavailableError(f'Pin {KONASHI_SPI_MISO_PINNB} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_SPI_MISO_PINNB].function]}')
            if self._gpio._config[KONASHI_SPI_MOSI_PINNB].function != int(GPIO.GPIOPinFunction.DISABLED) and self._gpio._config[KONASHI_SPI_MOSI_PINNB].function != int(GPIO.GPIOPinFunction.SPI):
                raise PinUnavailableError(f'Pin {KONASHI_SPI_MOSI_PINNB} is already configured as {GPIO._KONASHI_GPIO_FUNCTION_STR[self._gpio._config[KONASHI_SPI_MOSI_PINNB].function]}')
        b = bytearray([KONASHI_CFG_CMD_SPI]) + bytearray(config)
        await self._write(KONASHI_UUID_CONFIG_CMD, b)

    async def get_config(self) -> SPIConfig:
        """Get the current SPI configuration.

        Returns:
            SPIConfig: The SPI configuration.
        """
        await self._read(KONASHI_UUID_SPI_CONFIG_GET)
        return self._config

    async def transaction(self, write_data: bytes) -> bytes:
        """Perform an SPI transaction.

        Args:
            write_data (bytes): The data to send (length range is [1,127]).

        Raises:
            ValueError: The write data length is out of range.

        Returns:
            bytes: The received data (the length should be the same the write data length).
        """
        if len(write_data) == 0:
            raise ValueError("Write data buffer cannot be empty")
        if len(write_data) > 127:
            raise ValueError("Maximum write data length is 127 bytes")
        b = bytearray([KONASHI_CTL_CMD_SPI_DATA]) + bytearray(write_data)
        self._async_loop = asyncio.get_event_loop()
        self._data_in_future = self._async_loop.create_future()
        await self._write(KONASHI_UUID_CONTROL_CMD, b)
        res = await self._data_in_future
        self._async_loop = None
        self._data_in_future = None
        return res
